/*
 * @(#) XidImpl.java 
 *
 * JOTM: Java Open Transaction Manager 
 *
 *
 * This module was originally developed by 
 *
 *  - Bull S.A. as part of the JOnAS application server code released in 
 *    July 1999 (www.bull.com)
 * 
 * --------------------------------------------------------------------------
 *  The original code and portions created by Bull SA are 
 *  Copyright (c) 1999 BULL SA  
 *  All rights reserved.
 *  
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * --------------------------------------------------------------------------
 * $Id: XidImpl.java,v 1.17 2004/04/03 19:31:26 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.io.Serializable;
import java.nio.ByteBuffer;
import java.security.SecureRandom;

import javax.transaction.xa.Xid;

/**
 * Xid implementation for JTA
 *
 * XID has the following format as defined by X/Open Specification:
 *
 *     XID
 *         long formatID              format identifier
 *         long gtrid_length          value 1-64
 *         long bqual_length          value 1-64
 *         byte data [XIDDATASIZE]    where XIDDATASIZE = 128
 *
 *     The data field comprises at most two contiguous components:
 *     a global transaction identifier (gtrid) and a branch qualifier (bqual)
 *     which are defined as:
 *
 *         byte gtrid [1-64]          global transaction identfier
 *         byte bqual [1-64]          branch qualifier
 *
 */
public class XidImpl implements Xid, Serializable {

    public static final int JOTM_FORMAT_ID = 0xBB14;

	// these cells are gated by this.getClass()
	private static SecureRandom		rand = null;	// (also used as first-time flag)
	private final  byte				internalVersId = 1;
	private static int				count = 1;
	private static long				uuid0;
	private static long				uuid1;
	private static byte[]			gtrid_base = null;	// created by makeGtridBase()
	private static String			host, server;
	//

    /**
     * format id
     * @serial
     */
    private int formatId;

    /**
     * global transaction id
     * @serial
     */
    private byte[] gtrid;

    /**
     * branch qualifier
     * @serial
     */
    private byte[] bqual;

    // -------------------------------------------------------------------
    // Constructors
    // -------------------------------------------------------------------

    /**
     * Build an Xid for a local transaction
     */
    public XidImpl( String serverName, int ipAddr ) {

        TraceTm.jotm.debug("serverName=" + serverName + ", ipAddr=" + ipAddr);

        formatId = JOTM_FORMAT_ID;
        byte local_vers = 0;
        long uuid;
        long myipaddr;
        synchronized (getClass()) {
            uuid = System.currentTimeMillis() * 1024 + count;
            count++;
        }

        // compute gtrid
        // compute JOnAS gtrid
        String s =
            Long.toHexString(local_vers)
                + Long.toHexString(uuid)
                + "_"
                + Long.toHexString(ipAddr)
                + "_"
                + serverName;

        if (s.length() > Xid.MAXBQUALSIZE) {
            s = s.substring(0, Xid.MAXBQUALSIZE);
        }

        gtrid = s.getBytes();

        // XXX bqual is equal to gtrid in this constructor
        bqual = s.getBytes();

        TraceTm.jotm.debug("new Xid (uuid= " + Long.toHexString(uuid) + ")");

    }

    /**
     * Build an Xid from an otid_t CORBA
     */
    public XidImpl( int fid, int bqualsz, byte[] tid ) {
    	
        TraceTm.jotm.debug("constructor from otid_t");

        // FormatId
        formatId = fid;

        // gtrid
        int gtridsz = tid.length - bqualsz;
        gtrid = new byte[gtridsz];
        System.arraycopy(tid, bqualsz, gtrid, 0, gtridsz);

        // bqual
        bqual = new byte[bqualsz];
        System.arraycopy(tid, 0, bqual, 0, bqualsz);
    }

    /**
     * build Xid with all its components
     * (for performance on java serialization)
     */
    public XidImpl( int formatId, byte[] gtrid, byte[] bqual ) {
    	
        this.formatId = formatId;
        this.gtrid = gtrid;
        this.bqual = bqual;
    }

	/**
	 *	build a new Xid with a null BQual
	 */
	public XidImpl()
	{
		formatId = JOTM_FORMAT_ID;
		//
		gtrid = makeGtrid();
		//
		bqual = new byte[0];
 
	}

	/**
	 *	build a new Xid from an existing Xid, preserving the gtrid,
	 *	and using the 'index' as part of the unique but recognizable BQual.
	 *	The important thing is that this method must return the same result
	 *	on successive calls given the same input.
	 */
	public XidImpl( Xid oldXid, int index )
	{

                TraceTm.jotm.debug("old XID= " + oldXid);
                TraceTm.jotm.debug("index= " + index);
		formatId = oldXid.getFormatId();
		//
		gtrid = oldXid.getGlobalTransactionId();
		//
		makeGtridBase();
		bqual = new byte[MAXBQUALSIZE];
		ByteBuffer	bb = ByteBuffer.wrap(bqual);
		bb.put(gtrid_base);
		bb.putLong(0L);		// (time stamp not used on bqual)
		bb.putLong(index+0L);	// and this makes the bqual unique

	}

    // -------------------------------------------------------------------
    // Xid implementation
    // -------------------------------------------------------------------

    /**
     * Get the format id for that Xid
     */
    public int getFormatId() {
    	
        return formatId;
    }

    /**
     * Get the Global Id for that Xid
     */
    public byte[] getGlobalTransactionId() {
    	
        return gtrid;
    }

    /**
     * Get the Branch Qualifier for that Xid
     */
    public byte[] getBranchQualifier() {
    	
        return bqual;
    }

    // -------------------------------------------------------------------
    // other methods
    // -------------------------------------------------------------------

    /**
     * String form
     */

    public String toString() {

	byte[] my_gtrid_base = null;
        byte my_vers;
        long my_uuid0;
        long my_uuid1;
        long my_time;
        byte my_host [] = new byte [16];
        byte my_serv [] = new byte [15];
 
	byte[] bq_bqual_base = null;
        int bq_fmt;
        byte bq_vers;
        long bq_uuid0;
        long bq_uuid1;
        long bq_time;
        byte bq_host [] = new byte [16];
        byte bq_serv [] = new byte [15];

	my_gtrid_base = new byte[1+8+8+16+15+8];
	ByteBuffer aa = ByteBuffer.wrap(my_gtrid_base);
        System.arraycopy(gtrid, 0, my_gtrid_base, 0, gtrid.length);

	my_vers = aa.get();
	my_uuid0 = aa.getLong();
	my_uuid1 = aa.getLong();
	aa.get(my_host,0,16);
	aa.get(my_serv,0,15);
        my_time = aa.getLong();

        if (bqual.length == 0) {
            return Long.toHexString(formatId) +         // FORMAT ID
                   Long.toHexString(gtrid.length) +     // VALUE 1-64
                   Long.toHexString(bqual.length) +     // VALUE 1-64
                   Long.toHexString(my_vers) +          // INTERNAL VERSION ID
                   Long.toHexString(my_uuid0) +         // UNIQUE USERID PART 1
                   Long.toHexString(my_uuid1) +         // UNIQUE USERID PART 2
                   new String(my_host) +                // HOST NAME
                   new String(my_serv) +                // SERVER NAME
                   Long.toHexString(my_time) +          // TIMESTAMP
                   new String(bqual);
        } else {
            bq_bqual_base = new byte[1+8+8+16+15+8];
	    ByteBuffer bq = ByteBuffer.wrap(bq_bqual_base);

            if (bq_bqual_base.length > bqual.length) {
                System.arraycopy(bqual, 0, bq_bqual_base, 0, bqual.length);
            } else {
                System.arraycopy(bqual, 0, bq_bqual_base, 0, bq_bqual_base.length);
            }

	    bq_vers = bq.get();
	    bq_uuid0 = bq.getLong();
	    bq_uuid1 = bq.getLong();
	    bq.get(bq_host,0,16);
	    bq.get(bq_serv,0,15);
            bq_time = bq.getLong();

            return Long.toHexString(formatId) +         // FORMAT ID
                   Long.toHexString(gtrid.length) +     // VALUE 1-64
                   Long.toHexString(bqual.length) +     // VALUE 1-64
                   Long.toHexString(my_vers) +          // INTERNAL VERSION ID
                   Long.toHexString(my_uuid0) +         // UNIQUE USERID PART 1
                   Long.toHexString(my_uuid1) +         // UNIQUE USERID PART 2
                   new String(my_host) +                // HOST NAME
                   new String(my_serv) +                // SERVER NAME
                   Long.toHexString(my_time) +          // TIMESTAMP
                   Long.toHexString(bq_vers) +          // INTERNAL VERSION ID
                   Long.toHexString(bq_uuid0) +         // UNIQUE USERID PART 1
                   Long.toHexString(bq_uuid1) +         // UNIQUE USERID PART 2
                   new String(bq_host) +                // HOST NAME
                   new String(bq_serv) +                // SERVER NAME
                   Long.toHexString(bq_time);           // TIMESTAMP
        }
    }

	/*
	 *	make a unique but recognizable gtrid.
	 *
	 *	format:
	 *	1.	internal version identifier - 1 byte
	 *	2.	uuid                        - 16 bytes
	 *	3.	host name                   - 16 bytes
	 *	4.	server name                 - 15 bytes
	 *	5.	timestamp                   - 8 bytes
	 *	6.	[reserved for use by bqual  - 8 bytes]
	 *
	 *	Items 1 thru 4 are generated by makeGtridBase and comprise the recognizable
	 *	portion of a gtrid or bqual. Together, these serve to make Xids generated on
	 *	this system unique from those generated on other systems.
	 *	Item 5 is generated by this routine and serves to make gtrids unique.
	 *	Item 6 serves to distinguish different bquals belonging
	 *	to the same Xid, and is generated elsewhere.
	 *
	 *	Items 1 thru 4 are used to determine if we (JOTM) generated this gtrid/bqual.
	 */
	private byte[]  makeGtrid()
	{
		makeGtridBase();
		long	uniqueTimeStamp;
		synchronized( getClass() ) {
			uniqueTimeStamp = System.currentTimeMillis() * 1024 + count;
			count++;
		}
		ByteBuffer	bb = ByteBuffer.allocate(gtrid_base.length+8);
		bb.put(gtrid_base);
		bb.putLong(uniqueTimeStamp);
		return bb.array();
	}


	private void  makeGtridBase()
	{
		//
		// acquire the configured uuid, host and server name.
		// fabricate a uuid if one does not yet exist.
		// and append a unique timestamp.
		//
		synchronized( getClass() ) {
			if( rand == null ) {
				rand = new SecureRandom();
				//uuid0 = Long.parseLong(System.getProperty("jotm.uuid.part1"),16);
				//uuid1 = Long.parseLong(System.getProperty("jotm.uuid.part2"),16);
				uuid0 = rand.nextLong();
				uuid1 = rand.nextLong();
				//host = System.getProperty("jotm.thisHostName");
				//server = System.getProperty("jotm.thisServerName");
                                host = "";
                                server = "";
				// make host & server names fixed length, as defined above
				host = (host+"                ").substring(0,15);
				server = (server+"               ").substring(0,14);
				gtrid_base = new byte[1+8+8+16+15];
				ByteBuffer	bb = ByteBuffer.wrap(gtrid_base);
				bb.putLong(internalVersId);
				bb.putLong(uuid0);
				bb.putLong(uuid1);
				bb.put(host.getBytes());
				bb.put(server.getBytes());
			}
		}
	}


	/*
	 *	Check if the specified gtrid or bqual is one which was generated
	 *	by this version of JOTM.
	 *
	 *	@return true or false
	 */
	public boolean  IsThisOneOfOurs( byte[] gtrid_or_bqual )
	{
		if( rand == null ) { makeGtrid(); /* init gtrid_base if not yet done */ }
		boolean	ret = false;
		if( gtrid_or_bqual != null
		&&  gtrid_or_bqual.length >= gtrid_base.length ) {
			byte[] gbase;
			if( gtrid_or_bqual.length == gtrid_base.length ) {
				gbase = gtrid_base;
			} else {
				gbase = new byte[gtrid_base.length];
				System.arraycopy(gtrid_base,0,gbase,0,gtrid_base.length);
			}
			if( gtrid_or_bqual.equals(gbase) ) {
				ret = true;
			}
		}
		return ret;
	}


    // -------------------------------------------------------------------
    // equals and hashCode
    // -------------------------------------------------------------------

    /**
     * return true if objects are identical
     */
    public boolean equals( Object obj2 ) {
    	
        XidImpl xid2 = (XidImpl) obj2;

        if (formatId == xid2.getFormatId()
            && java.util.Arrays.equals(bqual, xid2.getBranchQualifier())
            && java.util.Arrays.equals(gtrid, xid2.getGlobalTransactionId())) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * return a hashcode value for this object
     */
    public int hashCode() {
    	
        String bquals = new String(bqual);
        String gtrids = new String(gtrid);
        return (gtrids + bquals).hashCode() % 100;
    }

}
