/*
 * @(#) TSHandler
 *
 * JOTM: Java Open Transaction Manager
 * 
 * This module was originally developed by 
 *  - INRIA inside the ObjectWeb Consortium(http://www.objectweb.org)
 *    
 * The original code and portions created by INRIA are 
 * Copyright (C) 2002 - INRIA (www.inria.fr)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * -Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. 
 *
 * -Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * --------------------------------------------------------------------------
 * $Id: TSHandler.java,v 1.1 2004/03/09 19:29:44 tonyortiz Exp $
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm.jta.jeremie;

import org.objectweb.david.apis.services.handler.Service;
import org.objectweb.jotm.TransactionContext;
import org.objectweb.jonathan.apis.kernel.Context;
import org.objectweb.jonathan.apis.kernel.JonathanException;
import org.objectweb.jonathan.apis.presentation.MarshallerFactory;
import org.objectweb.jonathan.apis.presentation.Marshaller;
import org.objectweb.jonathan.apis.presentation.UnMarshaller;
import org.objectweb.jonathan.apis.resources.Chunk;
import org.objectweb.jonathan.libs.helpers.MessageHelpers;
import org.omg.IOP.ServiceContext;

public class TSHandler implements Service {

    private JotmTransactionSender sender = null;
    private JotmTransactionReceiver receiver = null;
    private MarshallerFactory mf;
    private int service_id = 0;

   /** 
    * Builds a new Jeremie transaction service handler instance.
    * @param c unused
    * @param used_components the components used to initialize the new TSHandler.
    * @exception JonathanException if something goes wrong.
    */
   public TSHandler(Context c, Object[] used_components) throws JonathanException {
       int sid = ((Integer) used_components[0]).intValue();
       if (sid != Integer.MAX_VALUE) {
	   service_id = sid;
       }
       try {
	   sender = (JotmTransactionSender) used_components[1];
	   receiver = (JotmTransactionReceiver) used_components[2];
	   mf = (MarshallerFactory) used_components[3];
       } catch (Exception e) {
	   throw new JonathanException(e);
       }
   }
   
    /** 
     * Returns a request context.
     * @return                  a service context.
     */
    public ServiceContext getRequestContext(int id, boolean r, byte[] key, Context k) {
	if (sender == null) {
	    return null;
	}
	TransactionContext ctx = sender.sending_request();
	return encodeContext(ctx);
    }
	 
    
    /** 
     * Returns a reply context.
     * @return           a service context.
     */
    public ServiceContext getReplyContext(int id, Context k) {
	if (receiver == null) {
	    return null;
	}
	TransactionContext ctx = receiver.sending_reply();
	return encodeContext(ctx);
    }
    
    /** 
     * This method is called by the services handler to let the operations 
     * related to the target service be performed on request arrival.
     * @param context            the service context of the request;
     */
    public void handleRequestContext(ServiceContext context, int id, boolean r, byte[] key, Context k) {
	if (receiver == null || context == null) {
	    return;
	}
	TransactionContext ctx = decodeContext(context);
	if (ctx != null) {
	    receiver.received_request(ctx);
	}
    }
    
    /** 
     * This method is called by the services handler to let the operations 
     * related to the target service be performed on reply arrival.
     * @param context     the service context of the reply;
     */
    public void handleReplyContext(ServiceContext context, int id, Context k) {
	if (sender == null || context == null) {
	    return;
	}
	TransactionContext ctx = decodeContext(context);
	if (ctx != null) {
	    sender.received_reply(ctx);
	}
    }
    
   /**
    * Encodes a Transaction Service propagation context into an IOP service
    * context.
    * @param ctx the propagation context to be encoded.
    * @return the resulting IOP service context.
    */
   private ServiceContext encodeContext(TransactionContext ctx) {
      if (ctx == null) {
         return null;
      }
      Marshaller marshaller = mf.newMarshaller();
      byte[] byteArray = null;
      try {
         marshaller.writeValue(ctx);
         byteArray = MessageHelpers.copy(marshaller);
         marshaller.close();
      } catch (Exception e) {
         System.err.println("TSHandler.encodeContext: exception");
      }
      return new ServiceContext(service_id, byteArray);
   }
   
   /**
    * Decodes a Transaction Service propagation context from an IOP service
    * context.
    *
    * @param ctx the IOP service context containing the encoded propagation
    *    context.
    * @return the decoded propagation context.
    */
   private TransactionContext decodeContext(ServiceContext sc) {
      if (sc == null || sc.context_data == null || sc.context_data.length == 0 ) {
         return null;
      }
      TransactionContext ctx = null;
      byte[] byteArray = sc.context_data;
      UnMarshaller unmarshaller = 
         mf.newUnMarshaller(new Chunk(byteArray, 0, byteArray.length), 0);
      try {
         ctx = (TransactionContext) unmarshaller.readValue();
         unmarshaller.close();
      } catch (Exception e) {
         System.err.println("TSHandler.decodeContext: exception");
         System.err.println(e.toString() + "\n");
      }
      return ctx;
   }
   
}
